@@ -17,7 +17,12 @@
 #include <netinet/in.h> /* For htons */
 #include <netinet/ip_icmp.h>
 #include <libnetfilter_conntrack/libnetfilter_conntrack.h>
-#include <libnetfilter_conntrack/libnetfilter_conntrack_icmp.h>
+
+enum {
+	CT_ICMP_TYPE	= (1 << 0),
+	CT_ICMP_CODE	= (1 << 1),
+	CT_ICMP_ID	= (1 << 2),
+};
 
 static struct option opts[] = {
 	{"icmp-type", 1, 0, '1'},
@@ -67,41 +72,25 @@ static int parse(char c,
 		 unsigned int *flags)
 {
 	switch(c) {
+		u_int8_t tmp;
+		u_int16_t id;
 		case '1':
-			if (!optarg)
-				break;
-
-			nfct_set_attr_u8(ct, 
-					 ATTR_ICMP_TYPE,
-					 atoi(optarg));
-
+			tmp = atoi(optarg);
+			nfct_set_attr_u8(ct, ATTR_ICMP_TYPE, tmp);
 			nfct_set_attr_u8(ct, ATTR_L4PROTO, IPPROTO_ICMP);
-
-			*flags |= ICMP_TYPE;
+			*flags |= CT_ICMP_TYPE;
 			break;
 		case '2':
-			if (!optarg)
-				break;
-
-			nfct_set_attr_u8(ct, 
-					 ATTR_ICMP_CODE,
-					 atoi(optarg));
-
+			tmp = atoi(optarg);
+			nfct_set_attr_u8(ct, ATTR_ICMP_CODE, tmp);
 			nfct_set_attr_u8(ct, ATTR_L4PROTO, IPPROTO_ICMP);
-
-			*flags |= ICMP_CODE;
+			*flags |= CT_ICMP_CODE;
 			break;
 		case '3':
-			if (!optarg)
-				break;
-
-			nfct_set_attr_u16(ct,
-					  ATTR_ICMP_ID,
-					  htons(atoi(optarg)));
-
+			id = htons(atoi(optarg));
+			nfct_set_attr_u16(ct, ATTR_ICMP_ID, id);
 			nfct_set_attr_u8(ct, ATTR_L4PROTO, IPPROTO_ICMP);
-
-			*flags |= ICMP_ID;
+			*flags |= CT_ICMP_ID;
 			break;
 	}
 	return 1;
