@@ -59,8 +59,17 @@ enum action {
 
 	EXP_EVENT_BIT	= 14,
 	EXP_EVENT	= (1 << EXP_EVENT_BIT),
+
+	CT_COUNT_BIT	= 15,
+	CT_COUNT	= (1 << CT_COUNT_BIT),
+
+	EXP_COUNT_BIT	= 16,
+	EXP_COUNT	= (1 << EXP_COUNT_BIT),
+
+	X_STATS_BIT	= 17,
+	X_STATS		= (1 << X_STATS_BIT),
 };
-#define NUMBER_OF_CMD   15
+#define NUMBER_OF_CMD   18
 
 enum options {
 	CT_OPT_ORIG_SRC_BIT	= 0,
@@ -132,8 +141,11 @@ enum options {
 
 	CT_OPT_SECMARK_BIT	= 20,
 	CT_OPT_SECMARK		= (1 << CT_OPT_SECMARK_BIT),
+
+	CT_OPT_BUFFERSIZE_BIT	= 21,
+	CT_OPT_BUFFERSIZE	= (1 << CT_OPT_BUFFERSIZE_BIT),
 	
-	CT_OPT_MAX		= CT_OPT_SECMARK_BIT
+	CT_OPT_MAX		= CT_OPT_BUFFERSIZE_BIT
 };
 #define NUMBER_OF_OPT	CT_OPT_MAX+1
 
@@ -188,5 +200,6 @@ extern void register_tcp(void);
 extern void register_udp(void);
 extern void register_icmp(void);
 extern void register_icmpv6(void);
+extern void register_unknown(void);
 
 #endif
