@@ -4,6 +4,8 @@
 #include "mcast.h"
 #include "local.h"
 #include "alarm.h"
+#include "filter.h"
+#include "channel.h"
 
 #include <stdint.h>
 #include <stdio.h>
@@ -23,6 +25,15 @@
 #define REQUEST_DUMP	23	/* request dump 			*/
 #define DUMP_INT_XML	24	/* dump internal cache in XML		*/
 #define DUMP_EXT_XML	25	/* dump external cache in XML		*/
+#define RESET_TIMERS	26	/* reset kernel timers			*/
+#define DEBUG_INFO	27	/* unused				*/
+#define STATS_NETWORK	28	/* extended network stats		*/
+#define STATS_CACHE	29	/* extended cache stats			*/
+#define STATS_RUNTIME	30	/* extended runtime stats		*/
+#define STATS_LINK	31	/* dedicated link stats			*/
+#define STATS_QUEUE	32	/* queue stats				*/
+#define FLUSH_INT_CACHE	33	/* flush internal cache			*/
+#define FLUSH_EXT_CACHE	34	/* flush external cache			*/
 
 #define DEFAULT_CONFIGFILE	"/etc/conntrackd/conntrackd.conf"
 #define DEFAULT_LOCKFILE	"/var/lock/conntrackd.lock"
@@ -30,17 +41,6 @@
 #define DEFAULT_STATS_LOGFILE	"/var/log/conntrackd-stats.log"
 #define DEFAULT_SYSLOG_FACILITY	LOG_DAEMON
 
-enum {
-	SYNC_MODE_ALARM_BIT = 0,
-	SYNC_MODE_ALARM = (1 << SYNC_MODE_ALARM_BIT),
-
-	SYNC_MODE_FTFW_BIT = 1,
-	SYNC_MODE_FTFW = (1 << SYNC_MODE_FTFW_BIT),
-
-	DONT_CHECKSUM_BIT = 2,
-	DONT_CHECKSUM = (1 << DONT_CHECKSUM_BIT),
-};
-
 /* daemon/request modes */
 #define NOT_SET         0
 #define DAEMON		1
@@ -52,6 +52,7 @@ enum {
 #define CTD_SYNC_FTFW		(1UL << 2)
 #define CTD_SYNC_ALARM		(1UL << 3)
 #define CTD_SYNC_NOTRACK	(1UL << 4)
+#define CTD_POLL		(1UL << 5)
 
 /* FILENAME_MAX is 4096 on my system, perhaps too much? */
 #ifndef FILENAME_MAXLEN
@@ -71,23 +72,28 @@ struct ct_conf {
 	int syslog_facility;
 	char lockfile[FILENAME_MAXLEN];
 	int hashsize;			/* hashtable size */
-	struct mcast_conf mcast;	/* multicast settings */
+	int channel_num;
+	int channel_default;
+	int channel_type_global;
+	struct channel_conf channel[MULTICHANNEL_MAX];
 	struct local_conf local;	/* unix socket facilities */
+	int nice;
 	int limit;
 	int refresh;
 	int cache_timeout;		/* cache entries timeout */
 	int commit_timeout;		/* committed entries timeout */
-	int del_timeout;
+	unsigned int purge_timeout;	/* purge kernel entries timeout */
 	unsigned int netlink_buffer_size;
 	unsigned int netlink_buffer_size_max_grown;
-	unsigned char ignore_protocol[IPPROTO_MAX];
-	union inet_address *listen_to;
-	unsigned int listen_to_len;
+	int nl_overrun_resync;
 	unsigned int flags;
 	int family;			/* protocol family */
 	unsigned int resend_queue_size; /* FTFW protocol */
 	unsigned int window_size;
+	int poll_kernel_secs;
 	int cache_write_through;
+	int filter_from_kernelspace;
+	int event_iterations_limit;
 	struct {
 		char logfile[FILENAME_MAXLEN];
 		int syslog_facility;
@@ -103,19 +109,52 @@ struct ct_general_state {
 	FILE				*stats_log;
 	struct local_server		local;
 	struct ct_mode 			*mode;
-	struct ignore_pool		*ignore_pool;
+	struct ct_filter		*us_filter;
 
 	struct nfct_handle		*event;         /* event handler */
+	struct nfct_filter		*filter;	/* event filter */
+	int				event_iterations_limit;
+
 	struct nfct_handle		*dump;		/* dump handler */
-	struct nfct_handle		*overrun;	/* overrun handler */
-	struct alarm_block		overrun_alarm;
+	struct nfct_handle		*request;	/* request handler */
+	struct nfct_handle		*resync;	/* resync handler */
+	struct nfct_handle		*get;		/* get handler */
+	int				get_retval;	/* hackish */
+
+	struct alarm_block		resync_alarm;
+	struct alarm_block		polling_alarm;
 
 	struct fds			*fds;
 
 	/* statistics */
-	uint64_t			malformed;
-	uint64_t 			bytes[NFCT_DIR_MAX];
-	uint64_t 			packets[NFCT_DIR_MAX];
+	struct {
+		uint64_t 		bytes_orig;
+		uint64_t 		bytes_repl;
+		uint64_t 		packets_orig;
+		uint64_t 		packets_repl;
+
+		time_t			daemon_start_time;
+
+		uint64_t		nl_events_received;
+		uint64_t		nl_events_filtered;
+		uint32_t		nl_events_unknown_type;
+		uint32_t		nl_catch_event_failed;
+		uint32_t		nl_overrun;
+		uint32_t		nl_dump_unknown_type;
+		uint32_t		nl_kernel_table_flush;
+		uint32_t		nl_kernel_table_resync;
+
+		uint32_t		child_process_failed;
+		uint32_t		child_process_error_segfault;
+		uint32_t		child_process_error_term;
+
+		uint32_t		select_failed;
+		uint32_t		wait_failed;
+
+		uint32_t		local_read_failed;
+		uint32_t		local_unknown_request;
+
+	} stats;
 };
 
 #define STATE_SYNC(x) state.sync->x
@@ -124,16 +163,30 @@ struct ct_sync_state {
 	struct cache *internal; 	/* internal events cache (netlink) */
 	struct cache *external; 	/* external events cache (mcast) */
 
-	struct mcast_sock *mcast_server;  /* multicast socket: incoming */
-	struct mcast_sock *mcast_client;  /* multicast socket: outgoing  */
-	struct evfd *evfd;		  /* event fd */
+	struct multichannel	*channel;
+	struct nlif_handle	*interface;
+	struct queue *tx_queue;
+
+	struct alarm_block		reset_cache_alarm;
 
 	struct sync_mode *sync;		/* sync mode */
 
+	/* statistics */
+	struct {
+		uint64_t	msg_rcv_malformed;
+		uint32_t	msg_rcv_bad_version;
+		uint32_t	msg_rcv_bad_payload;
+		uint32_t	msg_rcv_bad_header;
+		uint32_t	msg_rcv_bad_type;
+		uint32_t	msg_rcv_truncated;
+		uint32_t	msg_rcv_bad_size;
+		uint32_t	msg_snd_malformed;
+		uint64_t	msg_rcv_lost;
+		uint64_t	msg_rcv_before;
+	} error;
+
 	uint32_t last_seq_sent;	/* last sequence number sent */
 	uint32_t last_seq_recv;	/* last sequence number recv */
-	uint64_t packets_replayed;	/* number of replayed packets */
-	uint64_t packets_lost;         /* lost packets: sequence tracking */
 };
 
 #define STATE_STATS(x) state.stats->x
@@ -159,14 +212,13 @@ extern struct ct_general_state st;
 
 struct ct_mode {
 	int (*init)(void);
-	int (*register_fds)(struct fds *fds);
 	void (*run)(fd_set *readfds);
 	int (*local)(int fd, int type, void *data);
 	void (*kill)(void);
 	void (*dump)(struct nf_conntrack *ct);
-	int (*overrun)(enum nf_conntrack_msg_type type,
-		       struct nf_conntrack *ct,
-		       void *data);
+	int (*resync)(enum nf_conntrack_msg_type type,
+		      struct nf_conntrack *ct,
+		      void *data);
 	int (*purge)(void);
 	void (*event_new)(struct nf_conntrack *ct);
 	void (*event_upd)(struct nf_conntrack *ct);
