@@ -53,27 +53,26 @@
 #endif
 #include <signal.h>
 #include <string.h>
+#include <netdb.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 #include <libnetfilter_conntrack/libnetfilter_conntrack.h>
 
-static const char cmdflags[NUMBER_OF_CMD]
-= {'L','I','U','D','G','F','E','V','h','L','I','D','G','F','E'};
-
-static const char cmd_need_param[NUMBER_OF_CMD]
-= { 2,  0,  0,  0,  0,  2,  2,  2,  2,  2,  0,  0,  0,  2,  2 };
-
 static const char *optflags[NUMBER_OF_OPT] = {
 "src","dst","reply-src","reply-dst","protonum","timeout","status","zero",
 "event-mask","tuple-src","tuple-dst","mask-src","mask-dst","nat-range","mark",
-"id","family","src-nat","dst-nat","output","secmark"};
+"id","family","src-nat","dst-nat","output","secmark","buffersize"};
 
 static struct option original_opts[] = {
 	{"dump", 2, 0, 'L'},
-	{"create", 1, 0, 'I'},
-	{"delete", 1, 0, 'D'},
-	{"update", 1, 0, 'U'},
-	{"get", 1, 0, 'G'},
-	{"flush", 1, 0, 'F'},
-	{"event", 1, 0, 'E'},
+	{"create", 2, 0, 'I'},
+	{"delete", 2, 0, 'D'},
+	{"update", 2, 0, 'U'},
+	{"get", 2, 0, 'G'},
+	{"flush", 2, 0, 'F'},
+	{"event", 2, 0, 'E'},
+	{"counter", 2, 0, 'C'},
+	{"stats", 0, 0, 'S'},
 	{"version", 0, 0, 'V'},
 	{"help", 0, 0, 'h'},
 	{"orig-src", 1, 0, 's'},
@@ -99,6 +98,7 @@ static struct option original_opts[] = {
 	{"src-nat", 2, 0, 'n'},
 	{"dst-nat", 2, 0, 'g'},
 	{"output", 1, 0, 'o'},
+	{"buffer-size", 1, 0, 'b'},
 	{0, 0, 0, 0}
 };
 
@@ -120,22 +120,25 @@ static unsigned int global_option_offset = 0;
 static char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 /* Well, it's better than "Re: Linux vs FreeBSD" */
 {
-          /*   s d r q p t u z e [ ] { } a m i f n g o c */
-/*CT_LIST*/   {2,2,2,2,2,0,2,2,0,0,0,0,0,0,2,0,2,2,2,2,2},
-/*CT_CREATE*/ {2,2,2,2,1,1,1,0,0,0,0,0,0,2,2,0,0,2,2,0,0},
-/*CT_UPDATE*/ {2,2,2,2,2,2,2,0,0,0,0,0,0,0,2,2,2,2,2,2,0},
-/*CT_DELETE*/ {2,2,2,2,2,2,2,0,0,0,0,0,0,0,2,2,2,2,2,2,0},
-/*CT_GET*/    {2,2,2,2,1,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,0},
-/*CT_FLUSH*/  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
-/*CT_EVENT*/  {2,2,2,2,2,0,0,0,2,0,0,0,0,0,2,0,0,2,2,2,2},
-/*VERSION*/   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
-/*HELP*/      {0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
-/*EXP_LIST*/  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0},
-/*EXP_CREATE*/{1,1,2,2,1,1,2,0,0,1,1,1,1,0,0,0,0,0,0,0,0},
-/*EXP_DELETE*/{1,1,2,2,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
-/*EXP_GET*/   {1,1,2,2,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
-/*EXP_FLUSH*/ {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
-/*EXP_EVENT*/ {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+          /*   s d r q p t u z e [ ] { } a m i f n g o c b*/
+/*CT_LIST*/   {2,2,2,2,2,0,2,2,0,0,0,0,0,0,2,0,2,2,2,2,2,0},
+/*CT_CREATE*/ {2,2,2,2,1,1,2,0,0,0,0,0,0,2,2,0,0,2,2,0,0,0},
+/*CT_UPDATE*/ {2,2,2,2,2,2,2,0,0,0,0,0,0,0,2,2,2,2,2,2,0,0},
+/*CT_DELETE*/ {2,2,2,2,2,2,2,0,0,0,0,0,0,0,2,2,2,2,2,2,0,0},
+/*CT_GET*/    {2,2,2,2,1,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,0,0},
+/*CT_FLUSH*/  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*CT_EVENT*/  {2,2,2,2,2,0,0,0,2,0,0,0,0,0,2,0,0,2,2,2,2,2},
+/*VERSION*/   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*HELP*/      {0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*EXP_LIST*/  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0},
+/*EXP_CREATE*/{1,1,2,2,1,1,2,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0},
+/*EXP_DELETE*/{1,1,2,2,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*EXP_GET*/   {1,1,2,2,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*EXP_FLUSH*/ {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*EXP_EVENT*/ {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*CT_COUNT*/  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*EXP_COUNT*/ {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
+/*X_STATS*/   {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
 };
 
 static LIST_HEAD(proto_list);
@@ -155,26 +158,61 @@ void register_proto(struct ctproto_handler *h)
 	list_add(&h->head, &proto_list);
 }
 
-static struct ctproto_handler *findproto(char *name)
+extern struct ctproto_handler ct_proto_unknown;
+
+static struct ctproto_handler *findproto(char *name, int *pnum)
 {
 	struct ctproto_handler *cur;
+	struct protoent *pent;
+	int protonum;
 
-	if (!name) 
-		return NULL;
-
+	/* is it in the list of supported protocol? */
 	list_for_each_entry(cur, &proto_list, head) {
-		if (strcmp(cur->name, name) == 0)
+		if (strcmp(cur->name, name) == 0) {
+			*pnum = cur->protonum;
 			return cur;
+		}
+	}
+	/* using the protocol name for an unsupported protocol? */
+	if ((pent = getprotobyname(name))) {
+		*pnum = pent->p_proto;
+		return &ct_proto_unknown;
+	}
+	/* using a protocol number? */
+	protonum = atoi(name);
+	if (protonum > 0 && protonum <= IPPROTO_MAX) {
+		/* try lookup by number, perhaps this protocol is supported */
+		list_for_each_entry(cur, &proto_list, head) {
+			if (cur->protonum == protonum) {
+				*pnum = protonum;
+				return cur;
+			}
+		}
+		*pnum = protonum;
+		return &ct_proto_unknown;
 	}
 
 	return NULL;
 }
 
 static void
-extension_help(struct ctproto_handler *h)
+extension_help(struct ctproto_handler *h, int protonum)
 {
-	fprintf(stdout, "\n");
-	fprintf(stdout, "Proto `%s' help:\n", h->name);
+	const char *name;
+
+	if (h == &ct_proto_unknown) {
+		struct protoent *pent;
+
+		pent = getprotobynumber(protonum);
+		if (!pent)
+			name = h->name;
+		else
+			name = pent->p_name;
+	} else {
+		name = h->name;
+	}
+
+	fprintf(stdout, "Proto `%s' help:\n", name);
 	h->help();
 }
 
@@ -186,20 +224,23 @@ exit_tryhelp(int status)
 	exit(status);
 }
 
-void __attribute__((noreturn))
-exit_error(enum exittype status, const char *msg, ...)
+static void free_options(void)
 {
-	va_list args;
-
-	/* On error paths, make sure that we don't leak the memory
-	 * reserved during options merging */
 	if (opts != original_opts) {
 		free(opts);
 		opts = original_opts;
 		global_option_offset = 0;
 	}
+}
+
+void __attribute__((noreturn))
+exit_error(enum exittype status, const char *msg, ...)
+{
+	va_list args;
+
+	free_options();
 	va_start(args, msg);
-	fprintf(stderr,"%s v%s: ", PROGNAME, VERSION);
+	fprintf(stderr,"%s v%s (conntrack-tools): ", PROGNAME, VERSION);
 	vfprintf(stderr, msg, args);
 	fprintf(stderr, "\n");
 	va_end(args);
@@ -208,15 +249,6 @@ exit_error(enum exittype status, const char *msg, ...)
 	exit(status);
 }
 
-static void
-generic_cmd_check(int command, int local_options)
-{
-	if (cmd_need_param[command] == 0 && !local_options)
-		exit_error(PARAMETER_PROBLEM,
-			   "You need to supply parameters to `-%c'",
-			   cmdflags[command]);
-}
-
 static int bit2cmd(int command)
 {
 	int i;
@@ -323,8 +355,8 @@ static struct parse_parameter {
 	size_t  size;
 	unsigned int value[6];
 } parse_array[PARSE_MAX] = {
-	{ {"ASSURED", "SEEN_REPLY", "UNSET", "FIXED_TIMEOUT"}, 4,
-	  { IPS_ASSURED, IPS_SEEN_REPLY, 0, IPS_FIXED_TIMEOUT} },
+	{ {"ASSURED", "SEEN_REPLY", "UNSET", "FIXED_TIMEOUT", "EXPECTED"}, 5,
+	  { IPS_ASSURED, IPS_SEEN_REPLY, 0, IPS_FIXED_TIMEOUT, IPS_EXPECTED} },
 	{ {"ALL", "NEW", "UPDATES", "DESTROY"}, 4,
 	  {~0U, NF_NETLINK_CONNTRACK_NEW, NF_NETLINK_CONNTRACK_UPDATE, 
 	   NF_NETLINK_CONNTRACK_DESTROY} },
@@ -342,12 +374,14 @@ do_parse_parameter(const char *str, size_t str_length, unsigned int *value,
 	struct parse_parameter *p = &parse_array[parse_type];
 
 	if (strncasecmp(str, "SRC_NAT", str_length) == 0) {
-		printf("skipping SRC_NAT, use --src-nat instead\n");
+		fprintf(stderr, "WARNING: ignoring SRC_NAT, "
+				"use --src-nat instead\n");
 		return 1;
 	}
 
 	if (strncasecmp(str, "DST_NAT", str_length) == 0) {
-		printf("skipping DST_NAT, use --dst-nat instead\n");
+		fprintf(stderr, "WARNING: ignoring DST_NAT, "
+				"use --dst-nat instead\n");
 		return 1;
 	}
 
@@ -382,7 +416,7 @@ static void
 add_command(unsigned int *cmd, const int newcmd)
 {
 	if (*cmd)
-		exit_error(PARAMETER_PROBLEM, "Invalid commands combination\n");
+		exit_error(PARAMETER_PROBLEM, "Invalid commands combination");
 	*cmd |= newcmd;
 }
 
@@ -403,9 +437,9 @@ check_type(int argc, char *argv[])
 	if (!table)
 		return 0;
 		
-	if (strncmp("expect", table, 6) == 0)
+	if (strncmp("expect", table, strlen(table)) == 0)
 		return 1;
-	else if (strncmp("conntrack", table, 9) == 0)
+	else if (strncmp("conntrack", table, strlen(table)) == 0)
 		return 0;
 	else
 		exit_error(PARAMETER_PROBLEM, "unknown type `%s'", table);
@@ -508,7 +542,9 @@ static const char usage_commands[] =
 	"  -I [table] parameters\t\tCreate a conntrack or expectation\n"
 	"  -U [table] parameters\t\tUpdate a conntrack\n"
 	"  -E [table] [options]\t\tShow events\n"
-	"  -F [table]\t\t\tFlush table\n";
+	"  -F [table]\t\t\tFlush table\n"
+	"  -C [table]\t\t\tShow counter\n"
+	"  -S\t\t\t\tShow statistics\n";
 
 static const char usage_tables[] =
 	"Tables: conntrack, expect\n";
@@ -540,6 +576,7 @@ static const char usage_parameters[] =
 	"  -f, --family proto\t\tLayer 3 Protocol, eg. 'ipv6'\n"
 	"  -t, --timeout timeout\t\tSet timeout\n"
 	"  -u, --status status\t\tSet status, eg. ASSURED\n"
+	"  -b, --buffer-size\t\tNetlink socket buffer size\n"
 	;
   
 
@@ -559,28 +596,12 @@ usage(char *prog)
 
 static unsigned int output_mask;
 
-static int ignore_nat(const struct nf_conntrack *obj,
-		      const struct nf_conntrack *ct)
+static int 
+filter_nat(const struct nf_conntrack *obj, const struct nf_conntrack *ct)
 {
 	uint32_t ip;
 
-	if (options & CT_OPT_SRC_NAT && options & CT_OPT_DST_NAT) {
-		if (!nfct_getobjopt(ct, NFCT_GOPT_IS_SNAT) &&
-		    !nfct_getobjopt(ct, NFCT_GOPT_IS_DNAT))
-			return 1;
-
-		if (nfct_attr_is_set(obj, ATTR_SNAT_IPV4)) {
-			ip = nfct_get_attr_u32(obj, ATTR_SNAT_IPV4);
-			if (ip != nfct_get_attr_u32(ct, ATTR_REPL_IPV4_DST))
-				return 1;
-		}
-
-		if (nfct_attr_is_set(obj, ATTR_DNAT_IPV4)) {
-			ip = nfct_get_attr_u32(obj, ATTR_DNAT_IPV4);
-			if (ip != nfct_get_attr_u32(ct, ATTR_REPL_IPV4_SRC))
-				return 1;
-		}
-	} else if (options & CT_OPT_SRC_NAT) {
+	if (options & CT_OPT_SRC_NAT) {
 		if (!nfct_getobjopt(ct, NFCT_GOPT_IS_SNAT))
 		  	return 1;
 
@@ -589,7 +610,8 @@ static int ignore_nat(const struct nf_conntrack *obj,
 			if (ip != nfct_get_attr_u32(ct, ATTR_REPL_IPV4_DST))
 				return 1;
 		}
-	} else if (options & CT_OPT_DST_NAT) {
+	}
+	if (options & CT_OPT_DST_NAT) {
 		if (!nfct_getobjopt(ct, NFCT_GOPT_IS_DNAT))
 			return 1;
 
@@ -604,11 +626,17 @@ static int ignore_nat(const struct nf_conntrack *obj,
 }
 
 static int counter;
+static int dump_xml_header_done = 1;
 
 static void __attribute__((noreturn))
 event_sighandler(int s)
 {
-	fprintf(stderr, "%s v%s: ", PROGNAME, VERSION);
+	if (dump_xml_header_done == 0) {
+		printf("</conntrack>\n");
+		fflush(stdout);
+	}
+
+	fprintf(stderr, "%s v%s (conntrack-tools): ", PROGNAME, VERSION);
 	fprintf(stderr, "%d flow events has been shown.\n", counter);
 	nfct_close(cth);
 	exit(0);
@@ -623,14 +651,21 @@ static int event_cb(enum nf_conntrack_msg_type type,
 	unsigned int op_type = NFCT_O_DEFAULT;
 	unsigned int op_flags = 0;
 
-	if (ignore_nat(obj, ct))
+	if (filter_nat(obj, ct))
 		return NFCT_CB_CONTINUE;
 
-	if (options & CT_COMPARISON && !nfct_cmp(obj, ct, NFCT_CMP_ALL))
+	if (options & CT_COMPARISON &&
+	    !nfct_cmp(obj, ct, NFCT_CMP_ALL | NFCT_CMP_MASK))
 		return NFCT_CB_CONTINUE;
 
-	if (output_mask & _O_XML)
+	if (output_mask & _O_XML) {
 		op_type = NFCT_O_XML;
+		if (dump_xml_header_done) {
+			dump_xml_header_done = 0;
+			printf("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
+			       "<conntrack>\n");
+		}
+	} 
 	if (output_mask & _O_EXT)
 		op_flags = NFCT_OF_SHOW_LAYER3;
 	if (output_mask & _O_TMS) {
@@ -644,7 +679,8 @@ static int event_cb(enum nf_conntrack_msg_type type,
 	if (output_mask & _O_ID)
 		op_flags |= NFCT_OF_ID;
 
-	nfct_snprintf(buf, 1024, ct, type, op_type, op_flags);
+	nfct_snprintf(buf, sizeof(buf), ct, type, op_type, op_flags);
+
 	printf("%s\n", buf);
 	fflush(stdout);
 
@@ -662,20 +698,27 @@ static int dump_cb(enum nf_conntrack_msg_type type,
 	unsigned int op_type = NFCT_O_DEFAULT;
 	unsigned int op_flags = 0;
 
-	if (ignore_nat(obj, ct))
+	if (filter_nat(obj, ct))
 		return NFCT_CB_CONTINUE;
 
-	if (options & CT_COMPARISON && !nfct_cmp(obj, ct, NFCT_CMP_ALL))
+	if (options & CT_COMPARISON &&
+	    !nfct_cmp(obj, ct, NFCT_CMP_ALL | NFCT_CMP_MASK))
 		return NFCT_CB_CONTINUE;
 
-	if (output_mask & _O_XML)
+	if (output_mask & _O_XML) {
 		op_type = NFCT_O_XML;
+		if (dump_xml_header_done) {
+			dump_xml_header_done = 0;
+			printf("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
+			       "<conntrack>\n");
+		}
+	}
 	if (output_mask & _O_EXT)
 		op_flags = NFCT_OF_SHOW_LAYER3;
 	if (output_mask & _O_ID)
 		op_flags |= NFCT_OF_ID;
 
-	nfct_snprintf(buf, 1024, ct, NFCT_T_UNKNOWN, op_type, op_flags);
+	nfct_snprintf(buf, sizeof(buf), ct, NFCT_T_UNKNOWN, op_type, op_flags);
 	printf("%s\n", buf);
 
 	counter++;
@@ -693,10 +736,11 @@ static int delete_cb(enum nf_conntrack_msg_type type,
 	unsigned int op_type = NFCT_O_DEFAULT;
 	unsigned int op_flags = 0;
 
-	if (ignore_nat(obj, ct))
+	if (filter_nat(obj, ct))
 		return NFCT_CB_CONTINUE;
 
-	if (options & CT_COMPARISON && !nfct_cmp(obj, ct, NFCT_CMP_ALL))
+	if (options & CT_COMPARISON &&
+	    !nfct_cmp(obj, ct, NFCT_CMP_ALL | NFCT_CMP_MASK))
 		return NFCT_CB_CONTINUE;
 
 	res = nfct_query(ith, NFCT_Q_DESTROY, ct);
@@ -712,7 +756,7 @@ static int delete_cb(enum nf_conntrack_msg_type type,
 	if (output_mask & _O_ID)
 		op_flags |= NFCT_OF_ID;
 
-	nfct_snprintf(buf, 1024, ct, NFCT_T_UNKNOWN, op_type, op_flags);
+	nfct_snprintf(buf, sizeof(buf), ct, NFCT_T_UNKNOWN, op_type, op_flags);
 	printf("%s\n", buf);
 
 	counter++;
@@ -735,7 +779,7 @@ static int print_cb(enum nf_conntrack_msg_type type,
 	if (output_mask & _O_ID)
 		op_flags |= NFCT_OF_ID;
 
-	nfct_snprintf(buf, 1024, ct, NFCT_T_UNKNOWN, op_type, op_flags);
+	nfct_snprintf(buf, sizeof(buf), ct, NFCT_T_UNKNOWN, op_type, op_flags);
 	printf("%s\n", buf);
 
 	return NFCT_CB_CONTINUE;
@@ -750,21 +794,22 @@ static int update_cb(enum nf_conntrack_msg_type type,
 	char __tmp[nfct_maxsize()];
 	struct nf_conntrack *tmp = (struct nf_conntrack *) (void *)__tmp;
 
-	memcpy(tmp, obj, sizeof(__tmp));
+	memset(tmp, 0, sizeof(__tmp));
 
-	if (ignore_nat(tmp, ct))
+	if (filter_nat(obj, ct))
 		return NFCT_CB_CONTINUE;
 
 	if (nfct_attr_is_set(obj, ATTR_ID) && nfct_attr_is_set(ct, ATTR_ID) &&
 	    nfct_get_attr_u32(obj, ATTR_ID) != nfct_get_attr_u32(ct, ATTR_ID))
 	    	return NFCT_CB_CONTINUE;
 
-	if (options & CT_OPT_TUPLE_ORIG && !nfct_cmp(tmp, ct, NFCT_CMP_ORIG))
+	if (options & CT_OPT_TUPLE_ORIG && !nfct_cmp(obj, ct, NFCT_CMP_ORIG))
 		return NFCT_CB_CONTINUE;
-	if (options & CT_OPT_TUPLE_REPL && !nfct_cmp(tmp, ct, NFCT_CMP_REPL))
+	if (options & CT_OPT_TUPLE_REPL && !nfct_cmp(obj, ct, NFCT_CMP_REPL))
 		return NFCT_CB_CONTINUE;
 
 	nfct_copy(tmp, ct, NFCT_CP_ORIG);
+	nfct_copy(tmp, obj, NFCT_CP_META);
 
 	res = nfct_query(ith, NFCT_Q_UPDATE, tmp);
 	if (res < 0)
@@ -800,12 +845,90 @@ static int dump_exp_cb(enum nf_conntrack_msg_type type,
 {
 	char buf[1024];
 
-	nfexp_snprintf(buf, 1024, exp, NFCT_T_UNKNOWN, NFCT_O_DEFAULT, 0);
+	nfexp_snprintf(buf,sizeof(buf), exp, NFCT_T_UNKNOWN, NFCT_O_DEFAULT, 0);
 	printf("%s\n", buf);
+	counter++;
 
 	return NFCT_CB_CONTINUE;
 }
 
+static int count_exp_cb(enum nf_conntrack_msg_type type,
+			struct nf_expect *exp,
+			void *data)
+{
+	counter++;
+	return NFCT_CB_CONTINUE;
+}
+
+#ifndef CT_STATS_PROC
+#define CT_STATS_PROC "/proc/net/stat/nf_conntrack"
+#endif
+
+/* As of 2.6.29, we have 16 entries, this is enough */
+#ifndef CT_STATS_ENTRIES_MAX
+#define CT_STATS_ENTRIES_MAX 64
+#endif
+
+/* maximum string length currently is 13 characters */
+#ifndef CT_STATS_STRING_MAX
+#define CT_STATS_STRING_MAX 64
+#endif
+
+static int display_proc_conntrack_stats(void)
+{
+	int ret = 0;
+	FILE *fd;
+	char buf[4096], *token, *nl;
+	char output[CT_STATS_ENTRIES_MAX][CT_STATS_STRING_MAX];
+	unsigned int value[CT_STATS_ENTRIES_MAX], i, max;
+
+	fd = fopen(CT_STATS_PROC, "r");
+	if (fd == NULL)
+		return -1;
+
+	if (fgets(buf, sizeof(buf), fd) == NULL) {
+		ret = -1;
+		goto out_err;
+	}
+
+	/* trim off trailing \n */
+	nl = strchr(buf, '\n');
+	if (nl != NULL) {
+		*nl = '\0';
+		nl = strchr(buf, '\n');
+	}
+	token = strtok(buf, " ");
+	for (i=0; token != NULL && i<CT_STATS_ENTRIES_MAX; i++) {
+		strncpy(output[i], token, CT_STATS_STRING_MAX);
+		output[i][CT_STATS_STRING_MAX-1]='\0';
+		token = strtok(NULL, " ");
+	}
+	max = i;
+
+	if (fgets(buf, sizeof(buf), fd) == NULL) {
+		ret = -1;
+		goto out_err;
+	}
+
+	nl = strchr(buf, '\n');
+	while (nl != NULL) {
+		*nl = '\0';
+		nl = strchr(buf, '\n');
+	}
+	token = strtok(buf, " ");
+	for (i=0; token != NULL && i<CT_STATS_ENTRIES_MAX; i++) { 
+		value[i] = (unsigned int) strtol(token, (char**) NULL, 16);
+		token = strtok(NULL, " ");
+	}
+
+	for (i=0; i<max; i++)
+		printf("%-10s\t\t%-8u\n", output[i], value[i]);
+
+out_err:
+	fclose(fd);
+	return ret;
+}
+
 static struct ctproto_handler *h;
 
 static const int cmd2type[][2] = {
@@ -817,6 +940,7 @@ static const int cmd2type[][2] = {
 	['E']	= { CT_EVENT,	EXP_EVENT },
 	['V']	= { CT_VERSION,	CT_VERSION },
 	['h']	= { CT_HELP,	CT_HELP },
+	['C']	= { CT_COUNT,	EXP_COUNT },
 };
 
 static const int opt2type[] = {
@@ -856,7 +980,7 @@ static const int opt2attr[] = {
 	['i']	= ATTR_ID,
 };
 
-static char exit_msg[][64] = {
+static char exit_msg[NUMBER_OF_CMD][64] = {
 	[CT_LIST_BIT] 		= "%d flow entries has been shown.\n",
 	[CT_CREATE_BIT]		= "%d flow entries has been created.\n",
 	[CT_UPDATE_BIT]		= "%d flow entries has been updated.\n",
@@ -872,6 +996,7 @@ int main(int argc, char *argv[])
 	int c, cmd;
 	unsigned int type = 0, event_mask = 0, l4flags = 0, status = 0;
 	int res = 0;
+	size_t socketbuffersize = 0;
 	int family = AF_UNSPEC;
 	char __obj[nfct_maxsize()];
 	char __exptuple[nfct_maxsize()];
@@ -882,7 +1007,7 @@ int main(int argc, char *argv[])
 	struct nf_conntrack *mask = (struct nf_conntrack *)(void*) __mask;
 	char __exp[nfexp_maxsize()];
 	struct nf_expect *exp = (struct nf_expect *)(void*) __exp;
-	int l3protonum;
+	int l3protonum, protonum = 0;
 	union ct_address ad;
 	unsigned int command = 0;
 
@@ -895,10 +1020,14 @@ int main(int argc, char *argv[])
 	register_udp();
 	register_icmp();
 	register_icmpv6();
+	register_unknown();
+
+	/* disable explicit missing arguments error output from getopt_long */
+	opterr = 0;
 
 	while ((c = getopt_long(argc, argv, "L::I::U::D::G::E::F::hVs:d:r:q:"
 					    "p:t:u:e:a:z[:]:{:}:m:i:f:o:n::"
-					    "g::c:", 
+					    "g::c:b:C::S", 
 					    opts, NULL)) != -1) {
 	switch(c) {
 		/* commands */
@@ -910,6 +1039,7 @@ int main(int argc, char *argv[])
 		case 'E':
 		case 'V':
 		case 'h':
+		case 'C':
 			type = check_type(argc, argv);
 			add_command(&command, cmd2type[c][type]);
 			break;
@@ -921,15 +1051,14 @@ int main(int argc, char *argv[])
 				exit_error(PARAMETER_PROBLEM, 
 					   "Can't update expectations");
 			break;
+		case 'S':
+			add_command(&command, X_STATS);
+			break;
 		/* options */
 		case 's':
 		case 'd':
 		case 'r':
 		case 'q':
-			if (!optarg)
-				exit_error(PARAMETER_PROBLEM, 
-					   "-%c requires an IP", c);
-
 			options |= opt2type[c];
 
 			l3protonum = parse_addr(optarg, &ad);
@@ -949,10 +1078,6 @@ int main(int argc, char *argv[])
 		case '}':
 		case '[':
 		case ']':
-			if (!optarg)
-				exit_error(PARAMETER_PROBLEM, 
-					   "-%c requires an IP", c);
-
 			options |= opt2type[c];
 			l3protonum = parse_addr(optarg, &ad);
 			set_family(&family, l3protonum);
@@ -968,12 +1093,8 @@ int main(int argc, char *argv[])
 			nfct_set_attr_u8(mask, ATTR_ORIG_L3PROTO, l3protonum);
 			break;
 		case 'p':
-			if (!optarg || !*optarg)
-				exit_error(PARAMETER_PROBLEM, 
-					   "-%c requires a valid protocol", c);
-
 			options |= CT_OPT_PROTO;
-			h = findproto(optarg);
+			h = findproto(optarg, &protonum);
 			if (!h)
 				exit_error(PARAMETER_PROBLEM,
 					   "`%s' unsupported protocol",
@@ -982,46 +1103,28 @@ int main(int argc, char *argv[])
 			opts = merge_options(opts, h->opts, &h->option_offset);
 			if (opts == NULL)
 				exit_error(OTHER_PROBLEM, "out of memory");
+
+			nfct_set_attr_u8(obj, ATTR_L4PROTO, protonum);
 			break;
 		case 't':
-			if (!optarg)
-				exit_error(PARAMETER_PROBLEM, 
-					   "-%c requires value", c);
-
 			options |= CT_OPT_TIMEOUT;
 			nfct_set_attr_u32(obj, ATTR_TIMEOUT, atol(optarg));
 			nfexp_set_attr_u32(exp, ATTR_EXP_TIMEOUT, atol(optarg));
 			break;
 		case 'u':
-			if (!optarg)
-				exit_error(PARAMETER_PROBLEM,
-					   "-%c requires type", c);
-
 			options |= CT_OPT_STATUS;
 			parse_parameter(optarg, &status, PARSE_STATUS);
 			nfct_set_attr_u32(obj, ATTR_STATUS, status);
 			break;
 		case 'e':
-			if (!optarg)
-				exit_error(PARAMETER_PROBLEM,
-					   "-%c requires type", c);
-
 			options |= CT_OPT_EVENT_MASK;
 			parse_parameter(optarg, &event_mask, PARSE_EVENT);
 			break;
 		case 'o':
-			if (!optarg)
-				exit_error(PARAMETER_PROBLEM,
-					   "-%c requires type", c);
-
 			options |= CT_OPT_OUTPUT;
 			parse_parameter(optarg, &output_mask, PARSE_OUTPUT);
 			break;
 		case 'z':
-			if (optarg)
-				exit_error(PARAMETER_PROBLEM,
-					   "-%c does not require parameters",c);
-
 			options |= CT_OPT_ZERO;
 			break;
 		case 'n':
@@ -1047,10 +1150,6 @@ int main(int argc, char *argv[])
 		case 'm':
 		case 'c':
 			options |= opt2type[c];
-			if (!optarg)
-				exit_error(PARAMETER_PROBLEM, 
-					   "-%c requires value", c);
-
 			nfct_set_attr_u32(obj,
 					  opt2attr[c],
 					  strtoul(optarg, NULL, 0));
@@ -1070,18 +1169,25 @@ int main(int argc, char *argv[])
 					   "`%s' unsupported protocol",
 					   optarg);
 			break;
+		case 'b':
+			socketbuffersize = atol(optarg);
+			options |= CT_OPT_BUFFERSIZE;
+			break;
+		case '?':
+			if (optopt)
+				exit_error(PARAMETER_PROBLEM,
+					   "option `%s' requires an "
+					   "argument", argv[optind-1]);
+			else
+				exit_error(PARAMETER_PROBLEM,
+					   "unknown option `%s'", 
+					   argv[optind-1]);
+			break;
 		default:
 			if (h && h->parse_opts 
 			    &&!h->parse_opts(c - h->option_offset, obj,
 			    		     exptuple, mask, &l4flags))
 				exit_error(PARAMETER_PROBLEM, "parse error");
-
-			/* Unknown argument... */
-			if (!h) {
-				usage(argv[0]);
-				exit_error(PARAMETER_PROBLEM,
-					   "Missing arguments...");
-			}
 			break;
 		}
 	}
@@ -1090,13 +1196,7 @@ int main(int argc, char *argv[])
 	if (family == AF_UNSPEC)
 		family = AF_INET;
 
-	/* set the protocol number if we have seen -p with no parameters */
-	if (h && !nfct_attr_is_set(obj, ATTR_ORIG_L4PROTO) &&
-	    !nfct_attr_is_set(obj, ATTR_REPL_L4PROTO))
-		nfct_set_attr_u8(obj, ATTR_L4PROTO, h->protonum);
-
 	cmd = bit2cmd(command);
-	generic_cmd_check(cmd, options);
 	generic_opt_check(options,
 			  NUMBER_OF_OPT,
 			  commands_v_options[cmd],
@@ -1129,6 +1229,11 @@ int main(int argc, char *argv[])
 		else
 			res = nfct_query(cth, NFCT_Q_DUMP, &family);
 
+		if (dump_xml_header_done == 0) {
+			printf("</conntrack>\n");
+			fflush(stdout);
+		}
+
 		nfct_close(cth);
 		break;
 
@@ -1237,7 +1342,7 @@ int main(int argc, char *argv[])
 			exit_error(OTHER_PROBLEM, "Can't open handler");
 		res = nfct_query(cth, NFCT_Q_FLUSH, &family);
 		nfct_close(cth);
-		fprintf(stderr, "%s v%s: ", PROGNAME, VERSION);
+		fprintf(stderr, "%s v%s (conntrack-tools): ",PROGNAME,VERSION);
 		fprintf(stderr,"connection tracking table has been emptied.\n");
 		break;
 
@@ -1257,10 +1362,28 @@ int main(int argc, char *argv[])
 
 		if (!cth)
 			exit_error(OTHER_PROBLEM, "Can't open handler");
+
+		if (options & CT_OPT_BUFFERSIZE) {
+			size_t ret;
+			ret = nfnl_rcvbufsiz(nfct_nfnlh(cth), socketbuffersize);
+			fprintf(stderr, "NOTICE: Netlink socket buffer size "
+					"has been set to %zu bytes.\n", ret);
+		}
 		signal(SIGINT, event_sighandler);
 		signal(SIGTERM, event_sighandler);
 		nfct_callback_register(cth, NFCT_T_ALL, event_cb, obj);
 		res = nfct_catch(cth);
+		if (res == -1) {
+			if (errno == ENOBUFS) {
+				fprintf(stderr, 
+					"WARNING: We have hit ENOBUFS! We "
+					"are losing events.\nThis message "
+					"means that the current netlink "
+					"socket buffer size is too small.\n"
+					"Please, check --buffer-size in "
+					"conntrack(8) manpage.\n");
+			}
+		}
 		nfct_close(cth);
 		break;
 
@@ -1274,32 +1397,59 @@ int main(int argc, char *argv[])
 		res = nfexp_catch(cth);
 		nfct_close(cth);
 		break;
-			
+	case CT_COUNT: {
+#define NF_CONNTRACK_COUNT_PROC "/proc/sys/net/netfilter/nf_conntrack_count"
+		int fd, count;
+		char buf[strlen("2147483647")];	/* INT_MAX */
+		fd = open(NF_CONNTRACK_COUNT_PROC, O_RDONLY);
+		if (fd == -1) {
+			exit_error(OTHER_PROBLEM, "Can't open %s",
+				   NF_CONNTRACK_COUNT_PROC);
+		}
+		if (read(fd, buf, sizeof(buf)) == -1) {
+			exit_error(OTHER_PROBLEM, "Can't read %s",
+				   NF_CONNTRACK_COUNT_PROC);
+		}
+		close(fd);
+		count = atoi(buf);
+		printf("%d\n", count);
+		break;
+	}
+	case EXP_COUNT:
+		cth = nfct_open(EXPECT, 0);
+		if (!cth)
+			exit_error(OTHER_PROBLEM, "Can't open handler");
+
+		nfexp_callback_register(cth, NFCT_T_ALL, count_exp_cb, NULL);
+		res = nfexp_query(cth, NFCT_Q_DUMP, &family);
+		nfct_close(cth);
+		printf("%d\n", counter);
+		break;
+	case X_STATS:
+		if (display_proc_conntrack_stats() < 0)
+			exit_error(OTHER_PROBLEM, "Can't open /proc interface");
+		break;
 	case CT_VERSION:
 		printf("%s v%s (conntrack-tools)\n", PROGNAME, VERSION);
 		break;
 	case CT_HELP:
 		usage(argv[0]);
 		if (options & CT_OPT_PROTO)
-			extension_help(h);
+			extension_help(h, protonum);
 		break;
 	default:
 		usage(argv[0]);
 		break;
 	}
 
-	if (opts != original_opts) {
-		free(opts);
-		opts = original_opts;
-		global_option_offset = 0;
-	}
-
 	if (res < 0)
 		exit_error(OTHER_PROBLEM, "Operation failed: %s",
 			   err2str(errno, command));
 
+	free_options();
+
 	if (exit_msg[cmd][0]) {
-		fprintf(stderr, "%s v%s: ", PROGNAME, VERSION);
+		fprintf(stderr, "%s v%s (conntrack-tools): ",PROGNAME,VERSION);
 		fprintf(stderr, exit_msg[cmd], counter);
 		if (counter == 0 && !(command & (CT_LIST | EXP_LIST)))
 			return EXIT_FAILURE;
