@@ -1,5 +1,5 @@
 /*
- * (C) 2006 by Pablo Neira Ayuso <pablo@netfilter.org>
+ * (C) 2006-2009 by Pablo Neira Ayuso <pablo@netfilter.org>
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,7 +19,6 @@
  */
 
 #include "mcast.h"
-#include "debug.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -29,6 +28,8 @@
 #include <sys/ioctl.h>
 #include <net/if.h>
 #include <errno.h>
+#include <limits.h>
+#include <libnfnetlink/libnfnetlink.h>
 
 struct mcast_sock *mcast_server_create(struct mcast_conf *conf)
 {
@@ -71,28 +72,12 @@ struct mcast_sock *mcast_server_create(struct mcast_conf *conf)
 	}
 
 	if ((m->fd = socket(conf->ipproto, SOCK_DGRAM, 0)) == -1) {
-		debug("mcast_sock_server_create:socket");
 		free(m);
 		return NULL;
 	}
 
-	if(conf->iface[0]) {
-		struct ifreq ifr;
-
-		strncpy(ifr.ifr_name, conf->iface, sizeof(ifr.ifr_name));
-
-		if (ioctl(m->fd, SIOCGIFMTU, &ifr) == -1) {
-			debug("ioctl");
-			close(m->fd);
-			free(m);
-			return NULL;
-		}
-		conf->mtu = ifr.ifr_mtu;
-	}
-
 	if (setsockopt(m->fd, SOL_SOCKET, SO_REUSEADDR, &yes, 
 				sizeof(int)) == -1) {
-		debug("mcast_sock_server_create:setsockopt1");
 		close(m->fd);
 		free(m);
 		return NULL;
@@ -107,7 +92,6 @@ struct mcast_sock *mcast_server_create(struct mcast_conf *conf)
 				sizeof(int)) == -1) {
 		/* not supported in linux kernel < 2.6.14 */
 		if (errno != ENOPROTOOPT) {
-			debug("mcast_sock_server_create:setsockopt2");
 			close(m->fd);
 			free(m);
 			return NULL;
@@ -117,7 +101,6 @@ struct mcast_sock *mcast_server_create(struct mcast_conf *conf)
 	getsockopt(m->fd, SOL_SOCKET, SO_RCVBUF, &conf->rcvbuf, &socklen);
 
 	if (bind(m->fd, (struct sockaddr *) &m->addr, m->sockaddr_len) == -1) {
-		debug("mcast_sock_server_create:bind");
 		close(m->fd);
 		free(m);
 		return NULL;
@@ -127,7 +110,6 @@ struct mcast_sock *mcast_server_create(struct mcast_conf *conf)
 	case AF_INET:
 		if (setsockopt(m->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
 			       &mreq.ipv4, sizeof(mreq.ipv4)) < 0) {
-			debug("mcast_sock_server_create:setsockopt2");
 			close(m->fd);
 			free(m);
 			return NULL;
@@ -136,7 +118,6 @@ struct mcast_sock *mcast_server_create(struct mcast_conf *conf)
 	case AF_INET6:
 		if (setsockopt(m->fd, IPPROTO_IPV6, IPV6_JOIN_GROUP,
 			       &mreq.ipv6, sizeof(mreq.ipv6)) < 0) {
-			debug("mcast_sock_server_create:setsockopt2");
 			close(m->fd);
 			free(m);
 			return NULL;
@@ -165,7 +146,6 @@ __mcast_client_create_ipv4(struct mcast_sock *m, struct mcast_conf *conf)
 
 	if (setsockopt(m->fd, IPPROTO_IP, IP_MULTICAST_LOOP, &no,
 		       sizeof(int)) < 0) {
-		debug("mcast_sock_client_create:setsockopt2");
 		close(m->fd);
 		return -1;
 	}
@@ -173,7 +153,6 @@ __mcast_client_create_ipv4(struct mcast_sock *m, struct mcast_conf *conf)
 	if (setsockopt(m->fd, IPPROTO_IP, IP_MULTICAST_IF,
 		       &conf->ifa.interface_addr,
 		       sizeof(struct in_addr)) == -1) {
-		debug("mcast_sock_client_create:setsockopt3");
 		close(m->fd);
 		return -1;
 	}
@@ -195,7 +174,6 @@ __mcast_client_create_ipv6(struct mcast_sock *m, struct mcast_conf *conf)
 
 	if (setsockopt(m->fd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &no,
 		       sizeof(int)) < 0) {
-		debug("mcast_sock_client_create:setsockopt2");
 		close(m->fd);
 		return -1;
 	}
@@ -203,7 +181,6 @@ __mcast_client_create_ipv6(struct mcast_sock *m, struct mcast_conf *conf)
 	if (setsockopt(m->fd, IPPROTO_IPV6, IPV6_MULTICAST_IF,
 		       &conf->ifa.interface_index6,
 		       sizeof(unsigned int)) == -1) {
-		debug("mcast_sock_client_create:setsockopt3");
 		close(m->fd);
 		return -1;
 	}
@@ -223,14 +200,12 @@ struct mcast_sock *mcast_client_create(struct mcast_conf *conf)
 	memset(m, 0, sizeof(struct mcast_sock));
 
 	if ((m->fd = socket(conf->ipproto, SOCK_DGRAM, 0)) == -1) {
-		debug("mcast_sock_client_create:socket");
 		free(m);
 		return NULL;
 	}
 
 	if (setsockopt(m->fd, SOL_SOCKET, SO_NO_CHECK, &conf->checksum, 
 				sizeof(int)) == -1) {
-		debug("mcast_sock_client_create:setsockopt1");
 		close(m->fd);
 		free(m);
 		return NULL;
@@ -245,7 +220,6 @@ struct mcast_sock *mcast_client_create(struct mcast_conf *conf)
 				sizeof(int)) == -1) {
 		/* not supported in linux kernel < 2.6.14 */
 		if (errno != ENOPROTOOPT) {
-			debug("mcast_sock_server_create:setsockopt2");
 			close(m->fd);
 			free(m);
 			return NULL;
@@ -281,7 +255,7 @@ void mcast_client_destroy(struct mcast_sock *m)
 	free(m);
 }
 
-ssize_t mcast_send(struct mcast_sock *m, void *data, int size)
+ssize_t mcast_send(struct mcast_sock *m, const void *data, int size)
 {
 	ssize_t ret;
 	
@@ -292,7 +266,6 @@ ssize_t mcast_send(struct mcast_sock *m, void *data, int size)
 		     (struct sockaddr *) &m->addr,
 		     m->sockaddr_len);
 	if (ret == -1) {
-		debug("mcast_sock_send");
 		m->stats.error++;
 		return ret;
 	}
@@ -315,7 +288,6 @@ ssize_t mcast_recv(struct mcast_sock *m, void *data, int size)
 		       (struct sockaddr *)&m->addr,
 		       &sin_size);
 	if (ret == -1) {
-		debug("mcast_sock_recv");
 		m->stats.error++;
 		return ret;
 	}
@@ -326,29 +298,55 @@ ssize_t mcast_recv(struct mcast_sock *m, void *data, int size)
 	return ret;
 }
 
-struct mcast_stats *mcast_get_stats(struct mcast_sock *m)
+int mcast_get_fd(struct mcast_sock *m)
 {
-	return &m->stats;
+	return m->fd;
 }
 
-void mcast_dump_stats(int fd, struct mcast_sock *s, struct mcast_sock *r)
+int
+mcast_snprintf_stats(char *buf, size_t buflen, char *ifname,
+		     struct mcast_stats *s, struct mcast_stats *r)
 {
-	char buf[512];
-	int size;
+	size_t size;
 
-	size = sprintf(buf, "multicast traffic:\n"
-			    "%20llu Bytes sent "
-			    "%20llu Bytes recv\n"
-			    "%20llu Pckts sent "
-			    "%20llu Pckts recv\n"
-			    "%20llu Error send "
-			    "%20llu Error recv\n\n",
-			    (unsigned long long)s->stats.bytes,
-			    (unsigned long long)r->stats.bytes,
-			    (unsigned long long)s->stats.messages,
-			    (unsigned long long)r->stats.messages,
-			    (unsigned long long)s->stats.error,
-			    (unsigned long long)r->stats.error);
+	size = snprintf(buf, buflen, "multicast traffic (active device=%s):\n"
+				     "%20llu Bytes sent "
+				     "%20llu Bytes recv\n"
+				     "%20llu Pckts sent "
+				     "%20llu Pckts recv\n"
+				     "%20llu Error send "
+				     "%20llu Error recv\n\n",
+				     ifname,
+				     (unsigned long long)s->bytes,
+				     (unsigned long long)r->bytes,
+				     (unsigned long long)s->messages,
+				     (unsigned long long)r->messages,
+				     (unsigned long long)s->error,
+				     (unsigned long long)r->error);
+	return size;
+}
 
-	send(fd, buf, size, 0);
+int
+mcast_snprintf_stats2(char *buf, size_t buflen, const char *ifname, 
+		      const char *status, int active,
+		      struct mcast_stats *s, struct mcast_stats *r)
+{
+	size_t size;
+
+	size = snprintf(buf, buflen, 
+			"multicast traffic device=%s status=%s role=%s:\n"
+			"%20llu Bytes sent "
+			"%20llu Bytes recv\n"
+			"%20llu Pckts sent "
+			"%20llu Pckts recv\n"
+			"%20llu Error send "
+			"%20llu Error recv\n\n",
+			ifname, status, active ? "ACTIVE" : "BACKUP",
+			(unsigned long long)s->bytes,
+			(unsigned long long)r->bytes,
+			(unsigned long long)s->messages,
+			(unsigned long long)r->messages,
+			(unsigned long long)s->error,
+			(unsigned long long)r->error);
+	return size;
 }
