@@ -18,11 +18,9 @@
 
 #include "netlink.h"
 #include "traffic_stats.h"
-#include "debug.h"
 #include "cache.h"
 #include "log.h"
 #include "conntrackd.h"
-#include "us-conntrack.h"
 
 #include <errno.h>
 #include <string.h>
@@ -38,9 +36,7 @@ static int init_stats(void)
 	}
 	memset(state.stats, 0, sizeof(struct ct_stats_state));
 
-	STATE_STATS(cache) = cache_create("stats",
-					  LIFETIME, 
-					  NULL); 
+	STATE_STATS(cache) = cache_create("stats", NO_FEATURES, NULL);
 	if (!STATE_STATS(cache)) {
 		dlog(LOG_ERR, "can't allocate memory for the "
 			      "external cache");
@@ -69,6 +65,7 @@ static int local_handler_stats(int fd, int type, void *data)
 		cache_dump(STATE_STATS(cache), fd, NFCT_O_XML);
 		break;
 	case FLUSH_CACHE:
+	case FLUSH_INT_CACHE:
 		dlog(LOG_NOTICE, "flushing caches");
 		cache_flush(STATE_STATS(cache));
 		break;
@@ -79,6 +76,9 @@ static int local_handler_stats(int fd, int type, void *data)
 		cache_stats(STATE_STATS(cache), fd);
 		dump_traffic_stats(fd);
 		break;
+	case STATS_CACHE:
+		cache_stats_extended(STATE_STATS(cache), fd);
+		break;
 	default:
 		ret = 0;
 		break;
@@ -96,15 +96,14 @@ static void dump_stats(struct nf_conntrack *ct)
 	nfct_attr_unset(ct, ATTR_TIMEOUT);
 	nfct_attr_unset(ct, ATTR_USE);
 
-	if (cache_update_force(STATE_STATS(cache), ct))
-		debug_ct(ct, "resync entry");
+	cache_update_force(STATE_STATS(cache), ct);
 }
 
-static int overrun_stats(enum nf_conntrack_msg_type type,
-			 struct nf_conntrack *ct,
-			 void *data)
+static int resync_stats(enum nf_conntrack_msg_type type,
+			struct nf_conntrack *ct,
+			void *data)
 {
-	if (ignore_conntrack(ct))
+	if (ct_filter_conntrack(ct, 1))
 		return NFCT_CB_CONTINUE;
 
 	/* This is required by kernels < 2.6.20 */
@@ -115,22 +114,21 @@ static int overrun_stats(enum nf_conntrack_msg_type type,
 	nfct_attr_unset(ct, ATTR_REPL_COUNTER_PACKETS);
 	nfct_attr_unset(ct, ATTR_USE);
 
-	if (!cache_test(STATE_STATS(cache), ct))
-		if (!cache_update_force(STATE_STATS(cache), ct))
-			debug_ct(ct, "overrun stats resync");
+	cache_update_force(STATE_STATS(cache), ct);
 
 	return NFCT_CB_CONTINUE;
 }
 
 static int purge_step(void *data1, void *data2)
 {
-	int ret;
-	struct us_conntrack *u = data2;
+	struct cache_object *obj = data2;
 
-	ret = nfct_query(STATE(dump), NFCT_Q_GET, u->ct);
-	if (ret == -1 && errno == ENOENT) {
-		debug_ct(u->ct, "overrun purge stats");
-		cache_del(STATE_STATS(cache), u->ct);
+	STATE(get_retval) = 0;
+	nl_get_conntrack(STATE(get), obj->ct); /* modifies STATE(get_retval) */
+	if (!STATE(get_retval)) {
+		cache_del(STATE_STATS(cache), obj);
+		dlog_ct(STATE(stats_log), obj->ct, NFCT_O_PLAIN);
+		cache_object_free(obj);
 	}
 
 	return 0;
@@ -145,52 +143,55 @@ static int purge_stats(void)
 
 static void event_new_stats(struct nf_conntrack *ct)
 {
+	int id;
+	struct cache_object *obj;
+
 	nfct_attr_unset(ct, ATTR_TIMEOUT);
 
-	if (cache_add(STATE_STATS(cache), ct)) {
-		debug_ct(ct, "cache new");
-	} else {
-		if (errno != EEXIST) {
-			dlog(LOG_ERR, 
-			     "can't add to cache cache: %s\n", strerror(errno));
-			debug_ct(ct, "can't add");
+	obj = cache_find(STATE_STATS(cache), ct, &id);
+	if (obj == NULL) {
+		obj = cache_object_new(STATE_STATS(cache), ct);
+		if (obj == NULL)
+			return;
+
+		if (cache_add(STATE_STATS(cache), obj, id) == -1) {
+			cache_object_free(obj);
+			return;
 		}
 	}
+	return;
 }
 
 static void event_update_stats(struct nf_conntrack *ct)
 {
 	nfct_attr_unset(ct, ATTR_TIMEOUT);
-
-	if (!cache_update_force(STATE_STATS(cache), ct)) {
-		debug_ct(ct, "can't update");
-		return;
-	}
-	debug_ct(ct, "update");
+	cache_update_force(STATE_STATS(cache), ct);
 }
 
 static int event_destroy_stats(struct nf_conntrack *ct)
 {
+	int id;
+	struct cache_object *obj;
+
 	nfct_attr_unset(ct, ATTR_TIMEOUT);
 
-	if (cache_del(STATE_STATS(cache), ct)) {
-		debug_ct(ct, "cache destroy");
+	obj = cache_find(STATE_STATS(cache), ct, &id);
+	if (obj) {
+		cache_del(STATE_STATS(cache), obj);
 		dlog_ct(STATE(stats_log), ct, NFCT_O_PLAIN);
+		cache_object_free(obj);
 		return 1;
-	} else {
-		debug_ct(ct, "can't destroy!");
-		return 0;
 	}
+	return 0;
 }
 
 struct ct_mode stats_mode = {
 	.init 			= init_stats,
-	.register_fds 		= NULL,
 	.run			= NULL,
 	.local			= local_handler_stats,
 	.kill			= kill_stats,
 	.dump			= dump_stats,
-	.overrun		= overrun_stats,
+	.resync			= resync_stats,
 	.purge			= purge_stats,
 	.event_new		= event_new_stats,
 	.event_upd		= event_update_stats,
